from __future__ import annotations

import attr
import pandas


class Node(object):
    def __init__(self, children=None):
        self.children = children


class RootNone(Node):
    def __init__(self):
        super().__init__()


class HierarchicalSeries(object):
    def __init__(self):
        ...




# -*- coding: utf-8 -*-
import logging

__author__ = """Carlo Mazzaferro"""
__email__ = 'carlo.mazzaferro@gmail.com'
__version__ = '0.1.0'

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

from hts.core.hierarchy import HierarchyTree
from hts.core.forecast import HierarchicalProphet
from hts.core.revision import CrossValidation, OLS, WLSS, WLSV, FP, AHP, PHA, BU



from enum import Enum

import numpy
import pandas
from fbprophet import Prophet

from hts import logger
from hts.core.functions import y_hat_matrix, optimal_combination, proportions, forecast_proportions


class RevisionMethod_(object):

    def __init__(self,
                 df,
                 sum_mat,
                 capacity,
                 capacity_future,
                 periods=1,
                 freq='D',
                 nodes=None,
                 # changepoints=None,
                 transformer=None,
                 # n_changepoints=None,
                 **kwargs):

        self.df = df
        self.sum_mat = sum_mat
        self.n_forecasts = self.sum_mat.shape[0]
        self.nodes = nodes
        self.capacity = capacity
        self.capacity_future = capacity_future
        # self.changepoints = changepoints
        # self.n_changepoints = n_changepoints
        self.periods = periods
        self.freq = freq
        self.kwargs = kwargs
        self.transformer = transformer
        self.models = dict()
        self.mse = dict()
        self.residuals = dict()
        self.forecasts = dict()

    def baseline_fit(self, df):
        for i, node in enumerate(self.df.columns[1:]):

            node_to_forecast = pandas.concat([df.iloc[:, [0]], df.iloc[:, i + 1]], axis=1)
            logger.info(f'Producing forecast for node: {node}')
            capacity = self.capacity.iloc[:, i] if self.capacity else None
            capacity_future = self.capacity_future.iloc[:, i] if self.capacity_future else None
            # changepoints = self.changepoints[:, i]
            # n_changepoints = self.n_changepoints[i]

            # Put the forecasts into a dictionary of dataframes
            # with contextlib.redirect_stdout(open(os.devnull, "w")):
            # Prophet related stuff
            node_to_forecast = node_to_forecast.rename(columns={node_to_forecast.columns[0]: 'ds'})
            node_to_forecast = node_to_forecast.rename(columns={node_to_forecast.columns[1]: 'y'})
            if self.capacity_future is None:
                growth = 'linear'
            else:
                growth = 'logistic'

            m = Prophet(growth,
                        # changepoints,
                        # n_changepoints,
                        **self.kwargs)
            node_to_forecast['capacity'] = capacity
            m.fit(node_to_forecast)

            inlc_hist = self.kwargs['include_history'] if 'include_history' in self.kwargs else True
            future = m.make_future_dataframe(periods=self.periods, freq=self.freq,
                                             include_history=inlc_hist)
            if self.capacity_future is not None:
                future['capacity'] = capacity_future

            # Base Forecasts, Residuals, and MSE
            self.forecasts[node] = m.predict(future)

            self.residuals[node] = df.iloc[:, i + 1] - self.forecasts[node].yhat[:-self.periods].values
            self.mse[node] = numpy.mean(numpy.array(self.residuals[node]) ** 2)
            self.models[node] = m
            # If logistic use exponential function, so that values can be added correctly
            if self.capacity_future is not None:
                self.forecasts[node].yhat = numpy.exp(self.forecasts[node].yhat)
        return self.forecasts

    def predict(self):
        raise NotImplementedError

    def _reformat(self, transformed):
        for i, key in enumerate(self.forecasts.keys()):
            values = transformed[:, i]
            self.forecasts[key].yhat = values

            # If Logistic fit values with natural log function to revert back to format of input
            if self.capacity_future is not None:
                self.forecasts[key].yhat = numpy.log(self.forecasts[key].yhat)
        return self.forecasts

    def _new_mat(self, y_hat_mat):
        new_mat = numpy.empty([y_hat_mat.shape[0], self.sum_mat.shape[0]])
        for i in range(y_hat_mat.shape[0]):
            new_mat[i, :] = numpy.dot(self.sum_mat, numpy.transpose(y_hat_mat[i, :]))
        return new_mat


class RevisionMethod(object):

    def __init__(self,
                 df,
                 forecasts,
                 mse,
                 sum_mat,
                 transformer,
                 ):
        self.df = df
        self.forecasts = forecasts
        self.transformer = transformer
        self.mse = mse
        self.sum_mat = sum_mat

    def _reformat(self, transformed):
        for i, key in enumerate(self.forecasts.keys()):
            values = transformed[:, i]
            self.forecasts[key].yhat = values

            # If Logistic fit values with natural log function to revert back to format of input
            if self.capacity_future is not None:
                self.forecasts[key].yhat = numpy.log(self.forecasts[key].yhat)
        return self.forecasts

    def _new_mat(self, y_hat_mat):
        new_mat = numpy.empty([y_hat_mat.shape[0], self.sum_mat.shape[0]])
        for i in range(y_hat_mat.shape[0]):
            new_mat[i, :] = numpy.dot(self.sum_mat, numpy.transpose(y_hat_mat[i, :]))
        return new_mat


class CrossValidation(RevisionMethod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def predict(self):
        pass


class OLS(RevisionMethod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def predict(self):
        if not self.forecasts:
            raise ValueError('Baseline must be fit first')

        transformed = optimal_combination(forecasts=self.forecasts,
                                          sum_mat=self.sum_mat,
                                          method='OLS',
                                          mse=self.mse)
        return self._reformat(transformed)


class WLSS(RevisionMethod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def predict(self):
        if not self.forecasts:
            raise ValueError('Baseline must be fit first')

        transformed = optimal_combination(forecasts=self.forecasts,
                                          sum_mat=self.sum_mat,
                                          method=Methods.WLSS.name,
                                          mse=self.mse)
        return self._reformat(transformed)


class WLSV(RevisionMethod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def predict(self):
        if not self.forecasts:
            raise ValueError('Baseline must be fit first')

        transformed = optimal_combination(forecasts=self.forecasts,
                                          sum_mat=self.sum_mat,
                                          method=Methods.WLSV.name,
                                          mse=self.mse)
        return self._reformat(transformed)


class BU(RevisionMethod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _y_hat_matrix(self):
        n_cols = len(list(self.forecasts.keys())) + 1
        keys = range(n_cols - self.sum_mat.shape[1] - 1, n_cols - 1)
        return y_hat_matrix(self.forecasts, keys=keys)

    def predict(self):
        if not self.forecasts:
            raise ValueError('Baseline must be fit first')
        y_hat = self._y_hat_matrix()
        return self._new_mat(y_hat)


class AHP(RevisionMethod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def predict(self):
        if not self.forecasts:
            raise ValueError('Baseline must be fit first')
        if self.transformer:
            for column in range(len(self.df.columns.tolist()) - 1):
                self.df.iloc[:, column + 1] = self.transformer.inverse_transform(self.df.iloc[:, column + 1])
        y_hat = proportions(self.df, self.forecasts, self.sum_mat, method=Methods.AHP.name)
        return self._new_mat(y_hat)


class PHA(RevisionMethod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def predict(self):
        if not self.forecasts:
            raise ValueError('Baseline must be fit first')
        if self.transformer:
            for column in range(len(self.df.columns.tolist()) - 1):
                self.df.iloc[:, column + 1] = self.transformer.inverse_transform(self.df.iloc[:, column + 1])
        y_hat = proportions(self.df, self.forecasts, self.sum_mat, method=Methods.PHA.name)
        return self._new_mat(y_hat)


class FP(RevisionMethod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def predict(self):
        if not self.forecasts:
            raise ValueError('Baseline must be fit first')
        return forecast_proportions(self.forecasts, self.nodes)


class Methods(Enum):
    CV = CrossValidation
    OLS = OLS
    WLSS = WLSS
    WLSV = WLSV
    FP = FP
    PHA = PHA
    AHP = AHP
    BU = BU



import pandas
from sklearn.base import BaseEstimator, TransformerMixin
from scipy.special import inv_boxcox
from scipy.stats import boxcox

from hts import logger


class FunctionTransformer(BaseEstimator, TransformerMixin):

    def __init__(self,
                 func: callable = boxcox,
                 inv_func: callable = inv_boxcox):

        self.func = func
        self.inv_func = inv_func
        self.lam = None

    def fit(self, X, y, **fit_params):
        return self

    def transform(self, X):
        try:
            for col in X.columns[1:]:
                X[col], self.lam = self.func(X[col])
        except RuntimeWarning as e:
            logger.error(f'Coulnd\'t perform transform: {repr(e)}')
        return X

    def fit_transform(self, X, y=None, **fit_params):
        try:
            for col in X.columns[1:]:
                X[col], self.lam = self.func(X[col], **fit_params)
        except RuntimeWarning as e:
            logger.error(f'Coulnd\'t perform transform: {repr(e)}')
        return X

    def inverse_transform(self, X):
        if isinstance(X, pandas.Series):
            return self.inv_func(X, self.lam)
        try:
            for col in X.columns[1:]:
                X[col] = self.inv_func(X[col], self.lam)
        except RuntimeWarning as e:
            logger.error(f'Coulnd\'t perform transform: {repr(e)}')
        return X







import numpy
import pandas


def partition_column(column, n=3):
    partitioned = column.apply(lambda x: numpy.random.dirichlet(numpy.ones(n),size=1).ravel() * x).values
    return [[i[j] for i in partitioned] for j in range(n)]


def hierarchical_sine_data(start, end, n=10000):
    dts = (end - start).total_seconds()
    dti = pandas.DatetimeIndex([start + pandas.Timedelta(numpy.random.uniform(0, dts), 's')
                                for _ in range(n)]).sort_values()
    time = numpy.arange(0, len(dti), 0.01)
    amplitude = numpy.sin(time) * 10
    amplitude += numpy.random.normal(2 * amplitude + 2, 5)
    df = pandas.DataFrame(index=dti, data={'total': amplitude[0:len(dti)]})
    df['a'], df['b'], df['c'] = partition_column(df.total, n=3)
    df['aa'], df['ab'] = partition_column(df.a, n=2)
    df['aaa'], df['aab'] = partition_column(df.aa, n=2)
    df['ba'], df['bb'], df['bc'] = partition_column(df.b, n=3)
    df['ca'], df['cb'], df['cc'], df['cd'] = partition_column(df.c, n=4)
    return df



from typing import Callable, Union, Optional
from typing_extensions import Literal
import pandas
from scipy.stats import boxcox
from scipy.special._ufuncs import inv_boxcox
from sklearn.base import BaseEstimator, RegressorMixin

from hts.core.factory import TimeSeriesModel
from hts.core.functions import to_sum_mat
from hts.core.revision import Methods
from hts.core.types import NAryTreeT, Transform, Models
from hts.transforms import FunctionTransformer


class HTS(BaseEstimator, RegressorMixin):

    def __init__(self,
                 nodes: NAryTreeT,
                 model: str = 'prophet',
                 periods: int = 1,
                 revision_method: str = 'OLS',
                 freq: str = 'D',
                 transform: Optional[Union[Transform, bool]] = None,
                 n_jobs: int = -1,
                 **kwargs
                 ):
        self.model = model
        self.periods = periods
        self.method = revision_method
        self.n_jobs = -1
        if transform is not None:
            if transform is True:
                self.transformer = FunctionTransformer(func=boxcox,
                                                       inv_func=inv_boxcox)
            else:
                self.transformer = FunctionTransformer(func=transform.func,
                                                       inv_func=transform.inv_func)
        else:
            self.transformer = FunctionTransformer(func=lambda x: (x, None),
                                                   inv_func=lambda x: (x, None))

        self.sum_mat = None
        self.nodes = nodes
        self.df = None
        self.revision_method = Methods[self.method].value(**kwargs)
        self.models = dict()
        self.mse = dict()
        self.residuals = dict()
        self.forecasts = dict()
        self.__init_hts()

    def _transform(self, df):
        time = df.columns[0]
        other = df.columns[1:]
        transf = self.transformer.transform(df[other])
        transf[time] = df[time]
        return df

    def _inv_transform(self, result):
        time = result.columns[0]
        other = result.columns[1:]
        transf = self.transformer.inverse_transform(result[other])
        transf[time] = result[time]
        return result

    def __init_hts(self):
        self.sum_mat = to_sum_mat(self.nodes)
        df = nodes.to_pandas().reset_index()
        self.df = self._transform(df)

    def fit(self):
        for i, node in [self.nodes] + self.nodes.traversal_level():
            model = TimeSeriesModel(kind=self.model, data=node)



class Forecast:

    def __init__(self, df: pandas.DataFrame):
        self.df = df

    @property
    def yhat(self):
        return self.df.yhat



from __future__ import annotations
import abc
import weakref
from enum import Enum
from typing import List, Optional, Callable, NamedTuple, NewType, Union

import pandas

from hts import logger

try:
    from folium import Map
except ImportError:
    logger.waring('Folium not installed, not all visualization will work')


class ExtendedEnum(Enum):

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def names(cls):
        return list(map(lambda c: c.name, cls))


class Model(ExtendedEnum):
    prophet = 'prophet'
    holt_winters = 'holt-winters'
    auto_arima = 'auto-arima'
    varmax = 'varmax'
    sarimax = 'sarimax'


class UnivariateModel(ExtendedEnum):
    arima = 'arima'
    auto_arima = 'auto-arima'
    prophet = 'prophet'
    holt_winters = 'holt-winters'


class MultivariateModel(ExtendedEnum):
    varmax = 'varmax'
    sarimax = 'sarimax'


Models = NewType('Model', Model)


class Transform(NamedTuple):
    func: Callable
    inv_func: Callable


class HierarchyVisualizerT(metaclass=abc.ABCMeta):
    tree: NAryTreeT

    def create_map(self) -> Map:
        ...


class NAryTreeT(metaclass=abc.ABCMeta):
    """
    Type definition of an NAryTree
    """
    key: str
    item: Union[pandas.Series, pandas.DataFrame]
    exogenous: List[str] = None
    children: List[Optional[NAryTreeT]]
    _parent: Optional[weakref.ref[NAryTreeT]]
    visualizer: HierarchyVisualizerT

    @property
    def parent(self):
        if self._parent:
            return self._parent()

    def __iter__(self) -> NAryTreeT:
        yield self
        for child in self.children:
            yield child

    def __getstate__(self):
        self._parent = None
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state
        for child in self.children:
            child._parent = weakref.ref(self)

    def __str__(self):
        return '{} : {}'.format(self.key, self.item)

    def num_nodes(self) -> int:
        ...

    def is_leaf(self) -> bool:
        ...

    def value_at_height(self, level: int) -> List:
        ...

    def sum_at_height(self, level) -> int:
        ...

    def get_height(self) -> int:
        ...

    def level_order_traversal(self: NAryTreeT) -> List[List[int]]:
        ...

    def traversal_level(self) -> List[NAryTreeT]:
        ...

    def add_child(self, key=None, item=None, exogenous=None) -> NAryTreeT:
        ...

    def leaf_sum(self) -> int:
        ...

    def to_pandas(self) -> pandas.DataFrame:
        ...

    def get_node_height(self, key: str) -> int:
        ...


m = Model.prophet.name
print(UnivariateModel.names())


import pandas
from h3 import h3

from hts.exceptions import InvalidArgumentException
from hts.core.types import NAryTreeT
from hts.core.hierarchy import HierarchyTree


def hexify(df, lat_col, lon_col, levels=(6, 8)):
    for r in range(levels[0], levels[1] + 1):
        df[f'hex_index_{r}'] = df.apply(lambda x: h3.geo_to_h3(x[lat_col], x[lon_col], r), 1)
    return df


def resample_count(df: pandas.DataFrame, freq: str, colname: str) -> pandas.DataFrame:
    _df = pandas.DataFrame({colname: df.index}, index=df.index)
    return _df.resample(freq).agg('count')


def groupify(df, freq='1H', nodes=None, min_count=0.1) -> NAryTreeT:
    """

    Parameters
    ----------
    df : pandas.DataFrame
    freq : str
        resample frequency
    nodes : tuple
        Hierarchy node
    min_count : int or float
        Minimum number of observations for a node to be used. If float, it represents a
        fraction of values that need to be non NaNs and must be between 0 and 1
    Returns
    -------

    """
    root_name = 'total'
    total = resample_count(df, freq, root_name)
    hierarchy = HierarchyTree(key=root_name, item=total)

    child_group = nodes[0]                    # city
    children = df[child_group].unique()       # berlin, munich, ...

    # add first level children
    for child in children:
        sub_df = df[df[child_group] == child]
        resampled = resample_count(sub_df, freq, child)
        hierarchy.add_child(key=child, item=resampled, exogenous=None)

    # add the rest
    for node in nodes[1:]:
        parent_group = child_group
        child_group = node                       # hex_index_6
        children = df[child_group].unique()      # abcccc, abccf, ...

        for child in children:
            sub_df = df[df[child_group] == child]
            if isinstance(min_count, float):
                allowance = len(total) * min_count
            elif isinstance(min_count, int):
                allowance = min_count
            else:
                raise InvalidArgumentException(f'min_count must be either float or integer')
            if len(sub_df) < allowance:
                continue
            parent_name = sub_df[parent_group].value_counts().index[0]
            resampled = resample_count(sub_df, freq, child)
            for c in hierarchy.traversal_level():
                if c.key == parent_name:
                    c.add_child(key=child, item=resampled, exogenous=None)
    return hierarchy



from __future__ import annotations

import weakref
from collections import deque
from itertools import chain
from typing import Tuple, Union, List, Optional

import pandas

from hts.core.types import NAryTreeT
from hts.viz.geo import HierarchyVisualizer


class HierarchyTree(NAryTreeT):
    """
    A generic N-ary tree implementations, that uses a list to store
    it's children.
    """

    @classmethod
    def from_events(cls,
                    df: pandas.DataFrame,
                    lat_col: str,
                    lon_col: str,
                    nodes: Tuple,
                    levels: Tuple[int, int] = (6, 7),
                    resample_freq: str = '1H',
                    min_count: Union[float, int] = 0.2,
                    ):

        from hts.helpers import hexify, groupify
        hexified = hexify(df, lat_col, lon_col, levels=levels)
        return groupify(hexified, nodes=nodes, freq=resample_freq, min_count=min_count)

    def __init__(self,
                 key: str = None,
                 item: Union[pandas.Series, pandas.DataFrame] = None,
                 exogenous: List[str] = None,
                 children: List[NAryTreeT] = None,
                 parent: NAryTreeT = None):

        self.key = key
        self.item = item
        if exogenous:
            self.exogenous = exogenous
        else:
            self.exogenous = []
        self.children = children or []
        self._parent = weakref.ref(parent) if parent else None
        self.visualizer = HierarchyVisualizer(self)

    def get_node(self, key: str) -> Optional[NAryTreeT]:
        for node in self.traversal_level():
            if node.key == key:
                return node
        return None

    def traversal_level(self) -> List[NAryTreeT]:
        res = []
        q = deque([(self, 0)])
        while q:
            n, li = q.popleft()
            if len(res) < li + 1:
                res.append([])
            for i in n.children:
                q.append((i, li + 1))
            res[li].extend(n.children)
        return list(chain.from_iterable(res[:-1]))

    def num_nodes(self) -> int:
        return sum(chain.from_iterable(self.level_order_traversal()))

    def is_leaf(self):
        return len(self.children) == 0

    def value_at_height(self, level: int) -> List:
        if level == 0:
            return [1]
        return self.level_order_traversal()[level - 1]

    def sum_at_height(self, level) -> int:
        return sum(self.value_at_height(level))

    def get_height(self) -> int:
        return len(self.level_order_traversal())

    def get_node_height(self, key: str) -> int:
        for node in self.traversal_level():
            if node.key == key:
                return node.get_height()
        return -1

    def level_order_traversal(self: NAryTreeT) -> List[List[int]]:
        res = []
        q = deque([(self, 0)])
        while q:
            n, li = q.popleft()
            if len(res) < li + 1:
                res.append([])
            for i in n.children:
                q.append((i, li + 1))
            res[li].append(len(n.children))
        return res[:-1]

    def add_child(self, key=None, item=None, exogenous=None) -> NAryTreeT:
        child = HierarchyTree(key=key, item=item, exogenous=exogenous, parent=self)
        self.children.append(child)
        return child

    def leaf_sum(self) -> int:
        return sum(self.level_order_traversal()[-1])

    def to_pandas(self):
        df = pandas.concat([self.item] + [c.item for c in self.traversal_level()], 1)
        df.index.name = 'ds'
        return df



import pandas

from hts import HierarchyTree
from hts.core.types import NAryTreeT


def test_level_order_traversal(n_tree):
    assert n_tree.level_order_traversal() == [[3], [2, 2, 2], [2, 2, 2, 2, 2, 2]]
    bottom = n_tree.get_node('aab')
    print(bottom)
    for i in bottom.level_order_traversal():
        assert i is None


def test_get_heights(n_tree):
    assert [sum(i) for i in n_tree.level_order_traversal()] == [3, 6, 12]


def test_get_node(n_tree):
    bab = n_tree.get_node('bab')
    assert isinstance(bab, NAryTreeT)
    assert bab.item == 6
    assert bab.is_leaf()
    assert n_tree.is_leaf() is False
    assert n_tree.leaf_sum() == 12
    assert n_tree.get_node('doesnotexist') is None


def test_create_from_events(events):
    ht = HierarchyTree.from_events(df=events,
                                   lat_col='start_latitude',
                                   lon_col='start_longitude',
                                   nodes=('city', 'hex_index_6', 'hex_index_7', 'hex_index_8'),
                                   levels=(6, 8),
                                   resample_freq='1H',
                                   min_count=0.5
                                   )
    assert isinstance(ht, NAryTreeT)
    assert len(ht.children) == events['city'].nunique()


def test_num_nodes(n_tree):
    assert n_tree.num_nodes() == 21


def test_height(n_tree):
    assert n_tree.value_at_height(1) == [3]
    assert n_tree.value_at_height(2) == [2, 2, 2]
    assert n_tree.value_at_height(3) == [2, 2, 2, 2, 2, 2]
    assert n_tree.sum_at_height(3) == 12
    assert n_tree.get_node_height('bb') == 1
    assert n_tree.get_node_height('bbb') == 0
    assert n_tree.get_node_height('a') == 2
    assert n_tree.get_node_height('z') == -1


def test_to_pandas(events):
    ht = HierarchyTree.from_events(df=events,
                                   lat_col='start_latitude',
                                   lon_col='start_longitude',
                                   nodes=('city', 'hex_index_6', 'hex_index_7', 'hex_index_8'),
                                   levels=(6, 8),
                                   resample_freq='1H',
                                   min_count=0.5
                                   )
    assert isinstance(ht.to_pandas(), pandas.DataFrame)


from typing import Dict

import numpy as np
import pandas

from hts.core.types import NAryTreeT


def to_sum_mat(ntree: NAryTreeT):
    """
    This function creates a summing matrix for the bottom up and optimal combination approaches
    All the inputs are the same as above
    The output is a summing matrix, see Rob Hyndman's "Forecasting: principles and practice" Section 9.4

    Parameters
    ----------
    ntree

    Returns
    -------

    """
    nodes = ntree.level_order_traversal()
    num_at_level = list(map(sum, nodes))
    columns = num_at_level[-1]
    bl_mat = np.identity(columns)
    top = np.ones(columns)
    final_mat = bl_mat
    num_levels = len(num_at_level)

    for lev in range(num_levels - 1):
        summing = nodes[-(lev + 1)]
        count = 0
        num2sum_ind = 0
        B = np.zeros([num_at_level[-1]])
        for num2sum in summing:
            num2sum_ind += num2sum
            a = bl_mat[count:num2sum_ind, :]
            count += num2sum
            if np.all(B == 0):
                B = a.sum(axis=0)
            else:
                B = np.vstack((B, a.sum(axis=0)))
        final_mat = np.vstack((B, final_mat))
        bl_mat = B

    final_mat = np.vstack((top, final_mat))
    return final_mat


def project(hat_mat: np.ndarray, sum_mat: np.ndarray, optimal_mat: np.ndarray) -> np.ndarray:
    new_mat = np.empty([hat_mat.shape[0], sum_mat.shape[0]])
    for i in range(hat_mat.shape[0]):
        new_mat[i, :] = np.dot(optimal_mat, np.transpose(hat_mat[i, :]))
    return new_mat


def y_hat_matrix(forecasts, keys=None):
    if not keys:
        keys = forecasts.keys()
    first = list(forecasts.keys())[0]
    y_hat_mat = np.zeros([len(forecasts[first].yhat), 1])
    for key in keys:
        f1 = np.array(forecasts[key].yhat)
        f2 = f1[:, np.newaxis]
        if np.all(y_hat_mat == 0):
            y_hat_mat = f2
        else:
            y_hat_mat = np.concatenate((y_hat_mat, f2), axis=1)
    return y_hat_mat


def optimal_combination(forecasts: Dict[str, pandas.DataFrame], sum_mat: np.ndarray, method: str, mse: Dict[str, float]):
    """
    Produces the optimal combination of forecasts by trace minimization (as described by
    Wickramasuriya, Athanasopoulos, Hyndman in "Optimal Forecast Reconciliation for Hierarchical and Grouped Time
    Series Through Trace Minimization")

    Parameters
    ----------
    forecasts : dict
        Dictionary of pandas.DataFrames containing the future predictions
    sum_mat : np.ndarray
        The summing  matrix
    method : str
        One of:
            - OLS (ordinary least squares)
            - WLSS (structurally weighted least squares)
            - WLSV (variance weighted least squares)
    mse

    Returns
    -------

    """
    hat_mat = y_hat_matrix(forecasts)
    transpose = np.transpose(sum_mat)

    if method == 'OLS':
        ols = np.dot(np.dot(sum_mat, np.linalg.inv(np.dot(transpose, sum_mat))), transpose)
        return project(hat_mat=hat_mat, sum_mat=sum_mat, optimal_mat=ols)
    elif method == 'WLSS':
        diag = np.diag(np.transpose(np.sum(sum_mat, axis=1)))
    elif method == 'WLSV':
        diag = [mse[key] for key in mse.keys()]
        diag = np.diag(np.flip(np.hstack(diag) + 0.0000001, 0))
    else:
        raise ValueError('Invalid method')

    # S*inv(S'S)*S'
    optimal_mat = np.dot(
        np.dot(np.dot(sum_mat, np.linalg.inv(np.dot(np.dot(transpose, np.linalg.inv(diag)), sum_mat))),
               transpose), np.linalg.inv(diag))

    return project(hat_mat=hat_mat, sum_mat=sum_mat, optimal_mat=optimal_mat)


def proportions(df, forecasts, sum_mat, method='APH'):
    n_cols = len(list(forecasts.keys())) + 1
    fcst = forecasts[0].yhat
    fcst = fcst[:, np.newaxis]
    num_bts = sum_mat.shape[1]
    bts_dat = pandas.DataFrame(df.iloc[:, n_cols - num_bts:n_cols])
    if method == 'APH':
        divs = np.divide(np.transpose(np.array(bts_dat)), np.array(df.iloc[:, 1]))
        props = divs.mean(1)
        props = props[:, np.newaxis]
    elif method == 'PHA':
        bts_sum = bts_dat.sum(0)
        top_sum = sum(df.iloc[:, 1])
        props = bts_sum / top_sum
        props = props[:, np.newaxis]
    else:
        raise ValueError('Invalid method')
    return np.dot(np.array(fcst), np.transpose(props))


def forecast_proportions(forecasts, nodes):
    """
    Cons:
       Produces biased revised forecasts even if base forecasts are unbiased
    """
    n_cols = len(list(forecasts.keys())) + 1
    ##
    # Find proportions of forecast at each step ahead, and then alter forecasts
    ##
    levels = len(nodes)
    column = 0
    first_node = 1
    new_mat = np.empty([len(forecasts[0].yhat), n_cols - 1])
    new_mat[:, 0] = forecasts[0].yhat
    for level in range(levels):
        nodes_in_level = len(nodes[level])
        for node in range(nodes_in_level):
            num_child = nodes[level][node]
            last_node = first_node + num_child
            lst = [x for x in range(first_node, last_node)]
            base_fcst = np.array([forecasts[k].yhat[:] for k in lst])
            fore_sum = np.sum(base_fcst, axis=0)
            fore_sum = fore_sum[:, np.newaxis]
            if column == 0:
                rev_top = np.array(forecasts[column].yhat)
                rev_top = rev_top[:, np.newaxis]
            else:
                rev_top = np.array(new_mat[:, column])
                rev_top = rev_top[:, np.newaxis]
            new_mat[:, first_node:last_node] = np.divide(np.multiply(np.transpose(base_fcst), rev_top), fore_sum)
            column += 1
            first_node += num_child
    return new_mat



from scipy.special._ufuncs import inv_boxcox
from scipy.stats import boxcox
from sklearn.base import BaseEstimator, RegressorMixin

from hts.core.functions import to_sum_mat
from hts.core.revision import Methods
from hts.transforms import FunctionTransformer
from hts.core.types import NAryTreeT


class HierarchicalProphet(BaseEstimator, RegressorMixin):

    """
    Parameters
    ----------
    periods
    nodes
    method
    freq
    transform
    include_history
    capacity
    capacity_future
    n_jobs
    kwargs
    """

    def __init__(self,
                 periods=1,
                 method='OLS',
                 freq='D',
                 transform=None,
                 include_history=True,
                 capacity=None,
                 capacity_future=None,
                 n_jobs=-1,
                 **kwargs):

        self.periods = periods
        self.freq = freq
        self.capacity = capacity
        self.capacity_future = capacity_future
        self.transform = transform
        if self.transform:
            if not isinstance(self.transform, dict):
                self.transformer = FunctionTransformer(func=boxcox,
                                                       inv_func=inv_boxcox)
            else:
                self.transformer = FunctionTransformer(transform)
        else:
            self.transformer = FunctionTransformer(func=lambda x: (x, None),
                                                   inv_func=lambda x: (x, None))
        self.include_history = include_history
        self.n_jobs = n_jobs
        self.method = method
        self.df = None
        self.sum_mat = None
        self.baseline = None
        self.model = None
        self.kwargs = kwargs

    @property
    def models(self):
        return self.model.models

    def _create_model(self, **kwargs):
        self.model = Methods[self.method].value(**kwargs)
        return self.model

    def _transform(self, df):
        time = df.columns[0]
        other = df.columns[1:]
        transf = self.transformer.transform(df[other])
        transf[time] = df[time]
        return df

    def _inv_transform(self, result):
        time = result.columns[0]
        other = result.columns[1:]
        transf = self.transformer.inverse_transform(result[other])
        transf[time] = result[time]
        return result

    def fit(self, nodes: NAryTreeT):
        self.sum_mat = to_sum_mat(nodes)
        df = nodes.to_pandas().reset_index()
        self.df = self._transform(df)

        self._create_model(df=self.df,
                           sum_mat=self.sum_mat,
                           capacity=self.capacity,
                           capacity_future=self.capacity_future,
                           periods=self.periods,
                           freq=self.freq,
                           nodes=nodes,
                           transformer=self.transformer,
                           **self.kwargs)

        self.baseline = self.model.baseline_fit(self.df)
        if self.transform:
            for i, node in range(self.model.n_forecasts):
                self.baseline[node].yhat = self.transformer.inverse_transform(self.baseline[node].yhat)
                self.baseline[node].trend = self.transformer.inverse_transform(self.baseline[node].trend)
                for component in ["seasonal", "daily", "weekly", "yearly", "holidays"]:
                    if component in self.baseline[node].columns.tolist():
                        inv_transf = self.transformer.inverse_transform(getattr(self.baseline[node], component))
                        setattr(self.baseline[node], component, inv_transf)

    def predict_future(self):
        return self.model.predict()



import os
import re

ROOT_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir)
DATA_DIR = os.path.join(ROOT_DIR, 'data')

ACCOUNT_ID = os.environ['AWS_ACCOUNT_ID']
TRAINING_JOB_NAME = os.environ['TRAINING_JOB_NAME']

INSTANCE_TYPE = 'ml.c4.8xlarge'

REGION_MAPPING = {'dev': 'us-east-2', 'test': 'us-west-1', 'prod': 'us-west-2', 'local': 'us-east-2'}

HAS_MODEL_REGEX = re.compile(r'{}-([0-9]\d*)-([0-9]\d*)-([0-9]\d*)\/output\/model.tar.gz'.format(TRAINING_JOB_NAME))
RUN_EXISTS_REGEX = re.compile(r'{}-([0-9]\d*)-([0-9]\d*)-([0-9]\d*)\/source\/sourcedir.tar.gz'.format(TRAINING_JOB_NAME))



from typing import Sequence, List, Union, Optional

import pandas
from fbprophet import Prophet
from pmdarima import AutoARIMA
from statsmodels.tsa.statespace.varmax import VARMAX
from statsmodels.tsa.statespace.sarimax import SARIMAX
from statsmodels.tsa.holtwinters import ExponentialSmoothing

from hts.core.types import Model, UnivariateModel, MultivariateModel, NAryTreeT
from hts.exceptions import InvalidArgumentException


class TimeSeriesModel:
    """
    Parameters
    ----------
    kind : str
        One of 'prophet', 'arima', 'auto-arima', 'varmax', 'holt-winters'
    data : pandas.DataFrame
        The data to which to fit the model
    endogenous : List[str]
        list of columns specifying the endogenous variables
    exogenous : Union[str, List[str]]
        Columns specifying the exogenous variables
    kwargs :
    """

    def __init__(self,
                 kind: str,
                 node: NAryTreeT,
                 endogenous: Union[str, List[str]] = 'y',
                 exogenous: Optional[List[str]] = None,
                 **kwargs):

        if kind not in Model.names():
            raise InvalidArgumentException(f'Model {kind} not valid. Pick one of: {" ".join(Model.names())}')

        if isinstance(endogenous, list):
            if kind in UnivariateModel.names():
                raise InvalidArgumentException(f'Multiple endogenous variables are allowed only for multivariate models:'
                                               f'{" ".join(MultivariateModel.names())}')
        self.kind = kind
        self.node = node
        self.model = self.create_model(**kwargs)
        self._add_exogenous_on_fit = False
        self._add_endogenous_on_fit = False

    def create_model(self, **kwargs):
        if self.kind == Model.prophet.name:
            self._add_endogenous_on_fit = False
            model = Prophet(**kwargs)
            if self.node.exogenous:
                for ex in self.node.exogenous:
                    model.add_regressor(ex)

        elif self.kind == Model.holt_winters.name:
            data = self.node.item
            model = ExponentialSmoothing(endog=data, **kwargs)

        elif self.kind == Model.arima.name:
            self._add_exogenous_on_fit = True
            self._add_endogenous_on_fit = True
            return AutoARIMA(**kwargs)

        elif self.kind == Model.varmax.name or self.kind == Model.sarimax.name:
            data = self.node.item[self.node.key]
            ex = self.node.item[self.node.exogenous]
            if self.kind == Model.varmax.name:
                return VARMAX(endog=data, exog=ex, **kwargs)
            else:
                return SARIMAX(endog=data, exog=ex, **kwargs)
        else:
            raise
        return model

    def fit(self, **kwargs):
        if self._add_endogenous_on_fit and self._add_exogenous_on_fit:
            end = self.node.item[self.node.key]
            ex = self.node.item[self.node.exogenous]
            return self.model.fit(y=end, exogenous=ex, **kwargs)
        if self._add_exogenous_on_fit:
            return self.model.fit(self.node.item, **kwargs)
        else:
            return self.model.fit(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from setuptools import setup, find_packages

with open('README.rst',  encoding='utf-8') as readme_file:
    readme = readme_file.read()

# Add here console scripts and other entry points in ini-AppBarStyle format
entry_points = """
[console_scripts]
script_name = hts.cli:main
"""


def strip_comments(l):
    return l.split('#', 1)[0].strip()


def _pip_requirement(req, *root):
    if req.startswith('-r '):
        _, path = req.split()
        return reqs(*root, *path.split('/'))
    return [req]


def _reqs(*f):
    path = (Path.cwd() / 'reqs').joinpath(*f)
    with path.open() as fh:
        reqs = [strip_comments(l) for l in fh.readlines()]
        return [_pip_requirement(r, *f[:-1]) for r in reqs if r]


def reqs(*f):
    return [req for subreq in _reqs(*f) for req in subreq]


install_requires = reqs('base.txt')
test_requires = reqs('test.txt') + install_requires

setup(
    author="Carlo Mazzaferro",
    author_email='carlo.mazzaferro@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Hierarchical Time Series forcasting",
    install_requires=install_requires,
    long_description=readme,
    include_package_data=True,
    keywords='scikit-hts',
    name='scikit-hts',
    packages=find_packages(include=['hts']),
    test_suite='tests',
    tests_require=test_requires,
    version='0.1.0',
    zip_safe=False,
)



import numpy as np
import pandas as pd

from hts.foreacast import HierarchicalProphet as HP

date = pd.date_range('2015-04-02', '2017-07-17')
date = np.repeat(date, 10)

country = ['Germany']
cities = ['Berlin', 'Hamburg']
district = ['Mitte', 'Kreuzberg', 'Fredechsein', 'Hamburg-Mitte', 'Altona']

district_rides = np.random.randint(1000, 10000, size=(len(date), len(district)))
b_city_rides = district_rides[:, :3].sum(axis=1)
h_city_rides = district_rides[:, 3:].sum(axis=1)
country_rides = district_rides.sum(axis=1)


data = {
    'time': date,
    'Total': country_rides,
    'Germany_Berlin': b_city_rides,
    'Germany_Hamburg': h_city_rides,
    'Germany_Berlin_Mitte': district_rides[:, 0],
    'Germany_Berlin_Kreuzberg': district_rides[:, 1],
    'Germany_Berlin_Fredechsein': district_rides[:, 2],
    'Germany_Hamburg_Hamburg-Mitte': district_rides[:, 3],
    'Germany_Hamburg_Altona': district_rides[:, 4],

}

df = pd.DataFrame(data)

holidates = pd.date_range('12/25/2013', '12/31/2017', freq='A')
holidays = pd.DataFrame(['Christmas'] * 5, columns=['holiday'])
holidays['ds'] = holidates
holidays['lower_window'] = [-4] * 5
holidays['upper_window'] = [0] * 5

nodes = [[2], [3, 2]]

if __name__ == '__main__':
    m = HP(periods=52, nodes=nodes, holidays=holidays, method='FP', transform=True)
    m.fit(df)
